#!/usr/bin/perl -w
use strict;
use warnings;

print STDERR "========================\nConvertFCodeTokensToForth.pl log:\n";

my ($fcodeInput, $hasPCIHeader) = @ARGV;

if (not defined $fcodeInput) {
	die "Need fcodeInput\n";
}

if (not defined $hasPCIHeader) {  # header info is in "/tmp/PCIHeader"
	$hasPCIHeader = 0;
	die "Need hasPCIHeader\n";
}

$_ = "";
my $FCode1 = "/tmp/FCode1.of";
my $FCode2 = "/tmp/FCode2.of";
my $FCode3 = "/tmp/FCode3.of";

{
	local $/=undef;
	open FILE, $fcodeInput or die "Couldn't open file: $fcodeInput";
	$_ = <FILE>;
	close FILE;
}

if (/[\cl\ck]/) {
	print STDERR "# File contains ctrl-j or ctrl-k characters.\n";
	exit 1; # make sure our special escape character (ctrl-k) doesnt exist}
}

# figure out what the indent character is

# assume indent character is tab
my $indentChar = " ";
my $indent = "    ";

# if exists a line that starts with a space character, then indent character is space, and a full indent is four spaces
if (/^[\t]/m) {
	$indentChar = "\t";
	$indent = "\t";
	#print STDERR "# Indent is tab character.\n";
} else {
	#print STDERR "# Indent is space character.\n";
}

my $indents = "[$indentChar]*";


# pci header

if ($hasPCIHeader != 0) {
	local $/=undef;
	open FILE, "/tmp/PCIHeader" or die "Couldn't open file: $fcodeInput";
	my $thePCIHeader = <FILE>;
	close FILE;
	substr($_, 0, 0, "$thePCIHeader");
}

# fcode-version2

my $gotstart = 0;
  s/start1 ((?:\\ \[0x0f1\] )?)\n  (format:    0x08)\n  (checksum:  0x[0-9a-f]+ \(.*?\))\n  (len:       0x[0-9a-f]+ \([0-9]+ bytes\))/fcode-version2 $1\n\\ $2\n\\ $3\n\\ $4\nhex\n\n/g && do { $gotstart = 1 };
s/version1 ((?:\\ \[0x0fd\] )?)\n  (format:    0x08)\n  (checksum:  0x[0-9a-f]+ \(.*?\))\n  (len:       0x[0-9a-f]+ \([0-9]+ bytes\))/fcode-version1 $1\n\\ $2\n\\ $3\n\\ $4\nhex\n\n/g && do { $gotstart = 1 };

if ($gotstart == 0) {
	print STDERR "# Didnt get start fcode.\n";
}

# fcode-end, pci-end

my $gotend = 0;
s/^([$indentChar]*)end0 (\\ \[0x000\] )?\n\\ detokenizing finished after .*?bytes\.\n/\n$1fcode-end\n/mg && do { $gotend = 1 };

if ($hasPCIHeader != 0) {
	$_ .= "pci-end\n";
}

if ($gotend == 0) {
	print STDERR "# Didnt get end fcode.\n";
	exit 1
}

my $missing = "";


# first do literal numbers so they can be used in token names

s/^([$indentChar]*)b\(lit\) ((?:\\ \[0x010\] )?)0x(?:ffffffff)?ffffffff\b/${1}-1 $2/mg || do { $missing .= "\nb(lit) ffffffff" }; # b(lit) ffffffff -> -1
s/^([$indentChar]*)b\(lit\) ((?:\\ \[0x010\] )?)0x(?:ffffffff)?([a-f][0-9a-f]{7})\b/${1}0$3 $2/mg || do { $missing .= "\nb(lit) a-fxxxxxxx" }; # b(lit) negative number needs a leading zero
s/^([$indentChar]*)b\(lit\) ((?:\\ \[0x010\] )?)0x(?:ffffffff)?([89][0-9a-f]{7})\b/${1}$3 $2/mg || do { $missing .= "\nb(lit) 8-9xxxxxxx" }; # b(lit) negative number
s/^([$indentChar]*)b\(lit\) ((?:\\ \[0x010\] )?)0x([a-f][0-9a-f]*)/${1}0$3 $2/mg || do { $missing .= "\nb(lit) a-f..." }; # b(lit) a-f...
s/^([$indentChar]*)b\(lit\) ((?:\\ \[0x010\] )?)0x([0-9][0-9a-f]*)/${1}$3 $2/mg || do { $missing .= "\nb(lit) 0-9..." }; # b(lit) 0-9...

# next do tokens

#	external        Newly created functions will be visible.                Arrange for subsequently created FCode functions to use external-token
#	headers         Newly created functions will be optionally visible.     Arrange for subsequently created FCode functions to use named-token
#	headerless      Newly created functions will be invisible.              Arrange for subsequently created FCode functions to use new-token
#	instance        value, variable, defer, or buffer:                      Used as: ok 30 instance value new-name
#
#	new-token       (Tokenized by defining words in headerless mode)        new-token fcode
#	named-token     (Tokenized by defining words in headers mode)           named-token name fcode
#	external-token  (Tokenized by defining words in external mode)          external-token name fcode
#
#
#	buffer:         new-token|named-token|external-token b(buffer:)         num <instance> buffer: name
#	value			new-token|named-token|external-token b(value)           num <instance> value name
#
#	defer			new-token|named-token|external-token b(defer)           <instance> defer name
#	variable		new-token|named-token|external-token b(variable)        <instance> variable name
#
#	constant        new-token|named-token|external-token b(constant)        num constant name
#	field           new-token|named-token|external-token b(field)           num field name
#
#	:               new-token|named-token|external-token b(:)				: name
#	create          new-token|named-token|external-token b(create)          create name


my @words = (
	"num <instance> buffer: name buffer",
	"num <instance> value name value",
	"<instance> defer name defer_word_function",
	"<instance> variable name variable",
	"num constant name const",
	"num field name field",
	": name colon_definition_function",
	"create name create_word_function"
);


foreach my $theword (@words) {
	if ($theword =~ /((?:num )*)((?:<instance> )*)(.*?) name (.*)/) {

		my $numpart = $1;
		my $instancepart = $2;
		my $fcodetype = $3;
		my $unnameprefix = $4;
		
		if ($instancepart) {
			if ($numpart) {
				s/^(?<r1>[$indentChar]*)(?<r2>[-]?[0-9][0-9a-f]*|\/[cwl]) (?<r6>(?:\\ \[0x[0-9a-f]{3}\] )?)\n(?<r4>(?:[$indentChar]*instance (?<r8>(\\ \[0x0c0\] )?)\n)*)[$indentChar]*new-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /$+{r4}\cknew\ck$+{r1}$+{r2} $fcodetype \cl${unnameprefix}_\cl$+{r3}_$+{r2}\cl \\ \($+{r3} $+{r2}\)\cl $+{r6}\ck$+{r7}/mg || do { $missing .= "\n\cknew\ck 1 $numpart$fcodetype" };
				s/^(?<r1>[$indentChar]*)(?<r2>[-]?[0-9][0-9a-f]*|\/[cwl]) (?<r6>(?:\\ \[0x[0-9a-f]{3}\] )?)\n(?<r4>(?:[$indentChar]*instance (?<r8>(\\ \[0x0c0\] )?)\n)*)[$indentChar]*named-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)(?<r5>[^ ]+) 0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /$+{r4}\cknamed\ck$+{r1}$+{r2} $fcodetype $+{r5} \\ \($+{r3} $+{r2}\) $+{r6}\ck$+{r7}/mg || do { $missing .= "\n\cknamed\ck 1 $numpart$fcodetype" };
				s/^(?<r1>[$indentChar]*)(?<r2>[-]?[0-9][0-9a-f]*|\/[cwl]) (?<r6>(?:\\ \[0x[0-9a-f]{3}\] )?)\n(?<r4>(?:[$indentChar]*instance (?<r8>(\\ \[0x0c0\] )?)\n)*)[$indentChar]*external-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)(?<r5>[^ ]+) 0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /$+{r4}\ckexternal\ck$+{r1}$+{r2} $fcodetype $+{r5} \\ \($+{r3} $+{r2}\) $+{r6}\ck$+{r7}/mg || do { $missing .= "\n\ckexternal\ck 1 $numpart$fcodetype" };
			}
			s/^(?<r1>[$indentChar]*)(?<r4>(?:[$indentChar]*instance (?<r8>(\\ \[0x0c0\] )?)\n)*)[$indentChar]*new-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /$+{r4}\cknew\ck$+{r1}$fcodetype \cl${unnameprefix}_\cl$+{r3}\cl \\ \($+{r3}\)\cl \ck$+{r7}/mg || do { $missing .= "\n\cknew\ck 2 $numpart$fcodetype" };
			s/^(?<r1>[$indentChar]*)(?<r4>(?:[$indentChar]*instance (?<r8>(\\ \[0x0c0\] )?)\n)*)[$indentChar]*named-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)(?<r5>[^ ]+) 0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /$+{r4}\cknamed\ck$+{r1}$fcodetype $+{r5} \\ \($+{r3}\) \ck$+{r7}/mg || do { $missing .= "\n\cknamed\ck 2 $numpart$fcodetype" };
			s/^(?<r1>[$indentChar]*)(?<r4>(?:[$indentChar]*instance (?<r8>(\\ \[0x0c0\] )?)\n)*)[$indentChar]*external-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)(?<r5>[^ ]+) 0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /$+{r4}\ckexternal\ck$+{r1}$fcodetype $+{r5} \\ \($+{r3}\) \ck$+{r7}/mg || do { $missing .= "\n\ckexternal\ck 2 $numpart$fcodetype" };
		}
		else {
			if ($numpart) {
				s/^(?<r1>[$indentChar]*)(?<r2>[-]?[0-9][0-9a-f]*|\/[cwl]) (?<r6>(?:\\ \[0x[0-9a-f]{3}\] )?)\n[$indentChar]*new-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /\cknew\ck$+{r1}$+{r2} $fcodetype \cl${unnameprefix}_\cl$+{r3}_$+{r2}\cl \\ \($+{r3} $+{r2}\)\cl $+{r6}\ck$+{r7}/mg || do { $missing .= "\n\cknew\ck 1 $numpart$fcodetype" };
				s/^(?<r1>[$indentChar]*)(?<r2>[-]?[0-9][0-9a-f]*|\/[cwl]) (?<r6>(?:\\ \[0x[0-9a-f]{3}\] )?)\n[$indentChar]*named-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)(?<r5>[^ ]+) 0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /\cknamed\ck$+{r1}$+{r2} $fcodetype $+{r5} \\ \($+{r3} $+{r2}\) $+{r6}\ck$+{r7}/mg || do { $missing .= "\n\cknamed\ck 1 $numpart$fcodetype" };
				s/^(?<r1>[$indentChar]*)(?<r2>[-]?[0-9][0-9a-f]*|\/[cwl]) (?<r6>(?:\\ \[0x[0-9a-f]{3}\] )?)\n[$indentChar]*external-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)(?<r5>[^ ]+) 0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /\ckexternal\ck$+{r1}$+{r2} $fcodetype $+{r5} \\ \($+{r3} $+{r2}\) $+{r6}\ck$+{r7}/mg || do { $missing .= "\n\ckexternal\ck 1 $numpart$fcodetype" };
			}
			s/^(?<r1>[$indentChar]*)[$indentChar]*new-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /\cknew\ck$+{r1}$fcodetype \cl${unnameprefix}_\cl$+{r3}\cl \\ \($+{r3}\)\cl \ck$+{r7}/mg || do { $missing .= "\n\cknew\ck 2 $numpart$fcodetype" };
			s/^(?<r1>[$indentChar]*)[$indentChar]*named-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)(?<r5>[^ ]+) 0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /\cknamed\ck$+{r1}$fcodetype $+{r5} \\ \($+{r3}\) \ck$+{r7}/mg || do { $missing .= "\n\cknamed\ck 2 $numpart$fcodetype" };
			s/^(?<r1>[$indentChar]*)[$indentChar]*external-token (?<r7>(\\ \[0x[0-9a-f]{3}\] )?)(?<r5>[^ ]+) 0x(?<r3>[0-9a-f]+)\n[$indentChar]*b\($fcodetype\) /\ckexternal\ck$+{r1}$fcodetype $+{r5} \\ \($+{r3}\) \ck$+{r7}/mg || do { $missing .= "\n\ckexternal\ck 2 $numpart$fcodetype" };
		}
	} else {
		exit 1;
	}
}

s/^[$indentChar]*instance ((?:\\ \[0x0c0] )?)\n(\ck[^\ck\n]+\ck[$indentChar]*(?:(?:[-0-9][0-9a-f]*|\/[cwl]) )?)([^\ck\n]+)/${2}instance $3$1/mg || do { $missing .= "\ninstance" }; # move instance to between num and forth word
s/^(\ck[^\ck\n]+\ck([$indentChar]*)(?:instance )?: )/$2\n$1/mg || do { $missing .= "\n:" }; # place a carriage return before :
s/^([$indentChar]*)0 ((?:\\ \[0x[0-9a-f]{3}\] )*)\n(\ck[^\ck\n]+\ck)([$indentChar]*(?:[0-9a-f]+|\/[cwl]) field )/$3$1struct $2\n$4/mg || do { $missing .= "\nstruct" }; # replace 0 with struct if it is located before a field instruction

# add headerless, headers and external compiler directives

my $nextToken = "";
my $replacement = "";

# /m makes ^ = start of line; /g is required so that the next search will start at pos instead of 0
while (/^(\ck(?<theToken>$nextToken[^\ck\n]+)\ck(?<tabs>[\t]*))/mg) {
	if ($+{theToken} eq "new") {
		$replacement = "$+{tabs}\n$+{tabs}headerless\n$+{tabs}";
		substr($_, $-[1], $+[1] - $-[1], $replacement);
		pos = $-[1] + length($replacement);
		$nextToken = "[ne][ax]";
	}
	elsif ($+{theToken} eq "named") {
		$replacement = "$+{tabs}\n$+{tabs}headers\n$+{tabs}";
		substr($_, $-[1], $+[1] - $-[1], $replacement);
		pos = $-[1] + length($replacement);
		$nextToken = "[ne][ex]";
	}
	elsif ($+{theToken} eq "external") {
		$replacement = "$+{tabs}\n$+{tabs}external\n$+{tabs}";
		substr($_, $-[1], $+[1] - $-[1], $replacement);
		pos = $-[1] + length($replacement);
		$nextToken = "n";
	}
}

s/^\ck[^\ck\n]+\ck//mg || do { $missing .= "\nnew, named, external" }; # new, named, external

s/\ck//g;


#s/([ \t\n])\(unnamed-fcode\)( \\ \[0x([0-9a-f]{3})\] )/$1unnamed-fcode_$3$2/g || do { $missing .= "\n(unnamed-fcode)" }; # unnamed-fcode

s/^(${indents})b\(\;\) ((?:\\ \[0x0c2\] )?)/$1; $2\n$1/mg || do { $missing .= "\nb(;)" }; # b(;) -> ;
s/^(${indents})b\(\'\) ((?:\\ \[0x011\] )?)([^ ]+ )((?:\\ \[0x[0-9a-f]{3}\] )?)\n/$1\[\'\] $3$2$4\n/mg || do { $missing .= "\nb(')" }; # b(') -> [']
s/^(${indents})b\(\"\) ((?:\\ \[0x012\] )?)(.*)/$1$3$2/mg || do { $missing .= "\nb(\")" }; # b(") -> "

if (0) {
	# lets not rename the new names to the old names
	s/^(${indents})lshift /$1<< /mg || do { $missing .= "\n<<" }; # lshift -> <<
	s/^(${indents})rshift /$1>> /mg || do { $missing .= "\n>>" }; # rshift -> >>
	s/^(${indents})chars /$1\/c* /mg || do { $missing .= "\n/c*" }; # chars -> /c*
	s/^(${indents})char\+ /$1ca1+ /mg || do { $missing .= "\nca1+" }; # char+ -> ca1+
	s/^(${indents})cells /$1\/n* /mg || do { $missing .= "\n/n*" }; # cells -> /n*
	s/^(${indents})cell\+ /$1na1+ /mg || do { $missing .= "\nna1+" }; # cell+ -> na1+
	s/^(${indents})invert /$1not /mg || do { $missing .= "\nnot" }; # invert -> not
	s/^(${indents})evaluate /$1eval /mg || do { $missing .= "\neval" }; # evaluate -> eval
}

s/^(${indents})\@ ((?:\\ \[0x06d\] )?)\n${indents}\. ((?:\\ \[0x09d\] )?)/$1? $2$3/mg || do { $missing .= "\n?" }; # @ . -> ?
s/^(${indents}[1-2]) ((?:\\ \[0x[0-9a-f]{3}\] )?)\n${indents}([-+]) ((?:\\ \[0x[0-9a-f]{3}\] )?)/$1$3 $2$4/mg || do { $missing .= "\n1+ 1- 2+ 2-" }; # 1 + 1 - 2 + 2 - -> 1+ 1- 2+ 2-
s/^(${indents})(\" .*?)\n${indents}type /$1.$2/mg || do { $missing .= "\n.\"" }; # " type -> ."
s/^(${indents})bl ((?:\\ \[0x0a9\] )?)\n${indents}fill ((?:\\ \[0x079\] )?)/$1blank $2$3/mg || do { $missing .= "\nblank" }; # bl fill -> blank

s/^(${indents})0a ((?:\\ \[0x[0-9a-f]{3}\] )?)\n${indents}base ((?:\\ \[0x0a0\] )?)\n${indents}! /$1decimal $2$3/mg || do { $missing .= "\ndecimal" }; # 0a base ! -> decimal
s/^(${indents})8 ((?:\\ \[0x[0-9a-f]{3}\] )?)\n${indents}base ((?:\\ \[0x0a0\] )?)\n${indents}! /$1octal $2$3/mg || do { $missing .= "\noctal" }; # 8 base ! -> octal
s/^(${indents})10 ((?:\\ \[0x[0-9a-f]{3}\] )?)\n${indents}base ((?:\\ \[0x0a0\] )?)\n${indents}! /$1hex $2$3/mg || do { $missing .= "\nhex" }; # 10 base ! -> hex

s/^(${indents})0 \\ \[0x0a5\] \n${indents}max \\ \[0x02f\] \n${indents}0 \\ \[0x0a5\] \n${indents}b\(\?do\) \\ \[0x018\] 0x7\n${indents}0 \\ \[0x0a5\] \n${indents}c, \\ \[0x0d0\] \n${indents}b\(loop\) \\ \[0x015\] 0x(?:ffffffff)?(?:ffff)?fffd\b/$1allot \\ [0x0a5] \\ [0x02f] \\ [0x0a5] \\ [0x018] \\ [0x0a5] \\ [0x0d0] \\ [0x015] /mg || do {
	s/^(${indents})0 \n${indents}max \n${indents}0 \n${indents}b\(\?do\) 0x7\n${indents}0 \n${indents}c, \n${indents}b\(loop\) 0x(?:ffffffff)?(?:ffff)?fffd\b/$1allot /mg || do { $missing .= "\nallot" }; # 0 max 0 b(?do) 0 c, b(loop) -> allot
};

s/^(${indents})span \\ \[0x088\] \n${indents}@ \\ \[0x06d\] \n${indents}-rot \\ \[0x04b\] \n${indents}expect \\ \[0x08a\] \n${indents}span \\ \[0x088\] \n${indents}@ \\ \[0x06d\] \n${indents}swap \\ \[0x049\] \n${indents}span \\ \[0x088\] \n${indents}! /$1accept \\ [0x088] \\ [0x06d] \\ [0x04b] \\ [0x08a] \\ [0x088] \\ [0x06d] \\ [0x049] \\ [0x088] /mg || do {
	s/^(${indents})span \n${indents}@ \n${indents}-rot \n${indents}expect \n${indents}span \n${indents}@ \n${indents}swap \n${indents}span \n${indents}! /$1accept /mg || do { $missing .= "\naccept" }; # span @ -rot expect span @ swap span ! -> accept
};

s/^(${indents})>r \\ \[0x030\] \n${indents}over \\ \[0x048\] \n${indents}r@ \\ \[0x032\] \n${indents}\+ \\ \[0x01e\] \n${indents}swap \\ \[0x049\] \n${indents}r@ \\ \[0x032\] \n${indents}- \\ \[0x01f\] \n${indents}rot \\ \[0x04A\] \n${indents}r> /$1decode-bytes \\ [0x030] \\ [0x048] \\ [0x032] \\ [0x01e] \\ [0x049] \\ [0x032] \\ [0x01f] \\ [0x04A] /mg || do {
	s/^(${indents})>r \n${indents}over \n${indents}r@ \n${indents}\+ \n${indents}swap \n${indents}r@ \n${indents}- \n${indents}rot \n${indents}r> /$1decode-bytes /mg || do { $missing .= "\ndecode-bytes" }; # >r over r@ + swap r@ - rot r> -> decode-bytes
};

s/^(${indents})base \\ \[0x0a0\] \n${indents}@ \\ \[0x06d\] \n${indents}swap \\ \[0x049\] \n${indents}decimal ((?:\\ \[0x[0-9a-f]{3}\] )*)\n${indents}. \\ \[0x09d\] \n${indents}base \\ \[0x0a0\] \n${indents}! \\ \[0x072\] /$1.d \\ [0x0a0] \\ [0x06d] \\ [0x049] $2\\ [0x09d] \\ [0x0a0] \\ [0x072] /mg || do {
	s/^(${indents})base \n${indents}@ \n${indents}swap \n${indents}decimal \n${indents}\. \n${indents}base \n${indents}! /$1.d/mg || do { $missing .= "\n.d" }; # base @ swap decimal . base ! -> .d
};

s/^(${indents})base \\ \[0x0a0\] \n${indents}@ \\ \[0x06d\] \n${indents}swap \\ \[0x049\] \n${indents}hex ((?:\\ \[0x[0-9a-f]{3}\] )*)\n${indents}. \\ \[0x09d\] \n${indents}base \\ \[0x0a0\] \n${indents}! \\ \[0x072\] /$1.h \\ [0x0a0] \\ [0x06d] \\ [0x049] $2\\ [0x09d] \\ [0x0a0] \\ [0x072] /mg || do {
	s/^(${indents})base \n${indents}@ \n${indents}swap \n${indents}hex \n${indents}\. \n${indents}base \n${indents}! /$1.h/mg || do { $missing .= "\n.h" }; # base @ swap hex . base ! -> .h
};

s/^(${indents})dup ((?:\\ \[0x047\] )?)\n${indents}abs ((?:\\ \[0x02d\] )?)\n${indents}<# ((?:\\ \[0x096\] )?)\n${indents}u#s ((?:\\ \[0x09a\] )?)\n${indents}swap ((?:\\ \[0x049\] )?)\n${indents}sign ((?:\\ \[0x098\] )?)\n${indents}u#> /$1(.) $2$3$4$5$6$7/mg || do { $missing .= "\n(.)" }; # dup abs <# u#s swap sign u#> -> (.)

s/^(${indents})2 ((?:\\ \[0x[0-9a-f]{3}\] )?)\n${indents}pick ((?:\\ \[0x04e\] )?)\n${indents}2 ((?:\\ \[0x[0-9a-f]{3}\] )?)\n${indents}pick ((?:\\ \[0x04e\] )?)\n${indents}2 ((?:\\ \[0x[0-9a-f]{3}\] )?)\n${indents}pick /${1}3dup $2$3$4$5$6/mg || do { $missing .= "\n3dup" }; # 2 pick 2 pick 2 pick -> 3dup
s/^(${indents})b\?branch ((?:\\ \[0x014\] )?)0x4\n${indents}b\(leave\) ((?:\\ \[0x01b\] )?)\n${indents}b\(\>resolve\) /$1\?leave $2$3/mg || do { $missing .= "\n?leave" }; # b?branch b(leave) b(>resolve) -> ?leave
s/^(${indents})bl ((?:\\ \[0x0a9\] )?)\n${indents}emit /$1space $2/mg || do { $missing .= "\nspace" }; # bl emit -> space
s/^(${indents})\(\.\) ((?:\\ \[0x[0-9a-f]{3}\] )*)\n${indents}type ((?:\\ \[0x090\] )?)\n${indents}space /$1s. $2$3/mg || do { $missing .= "\ns." }; # (.) type space -> s.

s/^(${indents})0 \\ \[0x0a5\] \n${indents}max \\ \[0x02f\] \n${indents}0 \\ \[0x0a5\] \n${indents}b\(\?do\) \\ \[0x018\] 0x7\n${indents}space \\ \[0x0a9\] \\ \[0x08f\] \n${indents}b\(loop\) \\ \[0x015\] 0x(?:ffffffff)?(?:ffff)?fffd\b/$1spaces \\ [0x0a5] \\ [0x02f] \\ [0x0a5] \\ [0x018] \\ [0x0a9] \\ [0x08f] \\ [0x015] /mg || do {
	s/^(${indents})0 \n${indents}max \n${indents}0 \n${indents}b\(\?do\) 0x7\n${indents}space \n${indents}b\(loop\) 0x(?:ffffffff)?(?:ffff)?fffd\b/$1spaces /mg || do { $missing .= "\nspaces" }; # 0 max 0 b(?do) space b(loop) -> spaces
};

s/^(${indents})<# ((?:\\ \[0x096\] )?)\n${indents}u#s ((?:\\ \[0x09a\] )?)\n${indents}u#> /$1(u.) $2$3/mg || do { $missing .= "\n(u.)" }; # <# u#s u#> -> (u.)
s/^(${indents})drop ((?:\\ \[0x046\] )?)\n${indents}2drop /${1}3drop $2/mg || do { $missing .= "\n3drop" }; # drop 2drop -> 3drop
s/^(${indents})0 ((?:\\ \[0x[0-9a-f]{3}\] )?)\n${indents}fill /$1erase $2/mg || do { $missing .= "\nerase" }; # 0 fill -> erase

s/^(${indents})b\(([a-z+?]+)\) ((?:\\ \[0x[0-9a-f]{3}\] )?)0x(?:ffffffff)?(?:ffff)?([89a-f][0-9a-f]{3})\b/$1$2 \\ \(0x$4\) $3/mg || do { $missing .= "\nb(loop,do,?do,endof,of) negative" }; # b(loop) b(+loop) b(do) b(?do) b(endof) b(of) negative offset
s/^(${indents})b\(([a-z+?]+)\) ((?:\\ \[0x[0-9a-f]{3}\] )?)(0x[0-9a-f]+)/$1$2 \\ \($4\) $3/mg || do { $missing .= "\nb(loop,do,?do,endof,of)" }; # b(loop) b(+loop) b(do) b(?do) b(endof) b(of)

s/^(${indents})b\(to\) ((?:\\ \[0x0c3\] )?)([^ ]+ )((?:\\ \[0x[0-9a-f]{3}\] )?)\n/${1}to $3$2$4\n/mg || do { $missing .= "\nb(to)" }; # b(to) -> to

s/^(${indents})b\(([a-z+?]+)\) /$1$2 /mg || do { $missing .= "\nb(leave,case,endcase)" }; # b(leave) b(case) b(endcase) -> leave case endcase

s/^(${indents})b\(<mark\) /$1begin /mg || do { $missing .= "\nbegin" }; # begin
s/^(${indents})b\?branch ((?:\\ \[0x014\] )?)0x(?:ffffffff)?(?:ffff)?([89a-f][0-9a-f]{3})\b/$1until \\ (0x$3) $2/mg || do { $missing .= "\nuntil" }; # until
s/^(${indents})${indent}bbranch ((?:\\ \[0x013\] )?)(0x[0-9a-f]{1,4})\n${indents}b\(>resolve\) /$1else \\ ($3) $2/mg || do { $missing .= "\nelse" }; # else
         s/^(${indents})bbranch ((?:\\ \[0x013\] )?)(0x[0-9a-f]{1,4})\n${indents}b\(>resolve\) /$1else \\ ($3) $2/mg && do { print STDERR "Got bad Else.\n" }; # else
s/^(${indents})bbranch ((?:\\ \[0x013\] )?)0x(?:ffffffff)?(?:ffff)?([89a-f][0-9a-f]{3})\b/$1again \\ (0x$3) $2/mg || do { $missing .= "\nagain" }; # again
s/^(${indents})b\(>resolve\) /$1then /mg || do { $missing .= "\nthen" }; # then
s/^(${indents})b\?branch ((?:\\ \[0x014\] )?)(0x[0-9a-f]{1,4})/$1if \\ ($3) $2/mg || do { $missing .= "\nif" }; # if

#  if ... again then (from previous transformations above) -> while ... repeat
my $repeats = 0;
#         1           2                                             3                                      4     5                                             
while (/^(${indents})(if) \\ \(0x[0-9a-f]{1,4}\) (?:\\ \[0x014\] )?(\n)(?:\1${indent}[^\n]*\n)*\1${indent}(again( \\ \(0x[89a-f][0-9a-f]{3}\) (?:\\ \[0x013\] )?)\n\1then )/mg) {
	substr($_, $-[4], $+[4] - $-[4], "repeat$5");
	substr($_, $-[2], $+[2] - $-[2], "while");
	pos = $-[3] + 3; # while is 3 characters longer than if
	$repeats++;
}
if ($repeats == 0) { $missing .= "\nwhile ... repeat" }

s/^(${indents}): \clcolon_definition_function_\cl([0-9a-f]{3})\cl( .*\n\1${indent}\" ([^ \n\"]*)\" .*\n\1${indent}; )/$1: \clcolon_string_\cl$2_$4\cl$3/mg || do { $missing .= "\ncolon_string" }; # colon_definition -> colon_string
s/^(${indents}): \clcolon_definition_function_\cl([0-9a-f]{3})\cl( .*\n\1${indent}([0-9][0-9a-f]*) .*\n\1${indent}; )/$1: \clcolon_const_\cl$2_$4\cl$3/mg || do { $missing .= "\ncolon_const" }; # colon_definition -> colon_const

# rename unnamed fcodes that we have given names to
my @fcodes;
while (/\cl([^\cl]+)\cl(([0-9a-f]{3})[^\cl]*)\cl([^\cl]*)\cl[^\n]*(\n)/g) {
	push @fcodes, [$+[4], $1.$2.$4, $3]
}
for my $fcode (reverse @fcodes) {
	pos = @$fcode[0];
	while ( /\G.*?\b(unnamed_fcode_@$fcode[2]) [^\n]*(\n)/mgs ) {
		substr($_, $-[1], $+[1] - $-[1], @$fcode[1]);
		pos = $+[2];
	}
}
s/\cl//g;

# outdent (then, until, again, loop, +loop, endcase, and repeat) but dont outdent else because it was done above
s/^(${indents})${indent}(then|until|again|endcase|endof|[+]?loop)\b/$1$2/mg || do { $missing .= "\nindent1" };
s/^(${indents})${indent}${indent}(repeat)\b/$1$2/mg || do { $missing .= "\nindent2" };

#===================================================
# Do formatting

# convert indent tabs to spaces
while (/^(\t+)/mg) {
	my $start = $-[1];
	my $length = $+[1] - $-[1];
	my $replacement = $1;
	$replacement =~ s/\t/    /g;
	substr($_, $start, $length, $replacement);
	pos = $start + length($replacement);
}

# protect slash definition
s/\\ \\/\ck \\/mg || do { $missing .= "\nslash definition" };

my $_s = '                                                                                '; # lots of spaces
while (/^(.+?)(\\[^\n["]*)([][()\\ 0-9a-fx-]*)$/mg) {
	my $start = $-[1];
	my $length = $+[3] - $-[1];

	# divide the line into the line part, comment part and instruction part.
	my $_l = $1; # line
	my $_c = $2; # comment
	my $_i = $3; # instruction

	# clean up line (remove trailing spaces)
	$_l =~ s/[ ]+$//;

	# clean up comment (remove trailing backslash and spaces)
	$_c =~ s/ \\ $//;
	$_c =~ s/[ ]+$//;

	#clean up the instructions (remove 0x and inner brackets)
	$_i =~ s/] \\ \[0x/ /g;
	$_i =~ s/\[0x/[/;

	# calculate tabs between comment and instructions
	my $_c2 = $_c.$_s;
	$_c2 =~ s/^(.{19} ) *$/$1/;			# comment that fits
	$_c2 =~ s/^(.{19}.*[^ ] ) *$/$1/;	# comment that doesn't fit
	$_c2 =~ s/^.*[^ ]( +)$/$1/;			# just grab the spaces
	$_c2 =~ s/    /\t/g;				# convert to tabs
	$_c2 =~ s/ +/\t/;					# convert remaining spaces to a tab

	# calculate tabs between the line and comment
	my $_l2 = $_l.$_s;
	$_l2 =~ s/^(.{79} ) *$/$1/;			# line that fits
	$_l2 =~ s/^(.{79}.*[^ ] ) *$/$1/;	# line that doesn't fits
	$_l2 =~ s/^.*[^ ]( +)$/$1/;			# just grab the spaces
	$_l2 =~ s/    /\t/g;				# convert to tabs
	$_l2 =~ s/ +/\t/;					# convert remaining spaces to a tab
	
	my $replacement = $_l.$_l2.$_c.$_c2.$_i;
	substr($_, $start, $length, $replacement);
	pos = $start + length($replacement) + 1;
}

# restore slash definition
s/\ck/\\/g;

# convert indent spaces to tabs
while (/^((?:    )+)/mg) {
	my $start = $-[1];
	my $length = $+[1] - $-[1];
	my $replacement = $1;
	$replacement =~ s/    /\t/g;
	substr($_, $start, $length, $replacement);
	pos = $start + length($replacement);
}

# remove trailing spaces and tabs
s/[ \t]+$//mg;


#===================================================

# dump debug stuff
while (/^(.*?unnamed_fcode_.*\n)/mg) {
	print STDERR $1;
}

print STDERR "\nThe following were not encountered:\n".$missing."\n";

print STDOUT $_;

exit 0;


=begin GHOSTCODE

" quote
stack: ( [text<">< >] -- text-str text-len )
code: none
generates: b(") len-byte xx-byte  xx-byte
Gathers the immediately following text string or hex data until reaching the terminator
"<whitespace>.


." dot quote
stack: ( [text<">] -- )
code: none
generates: b(") len text type
This word compiles a text string, delimited by "<whitespace> e.g. ." hello
world" .
At execution time, the string is displayed. This word is equivalent to using " text"
type .
." is normally used only within a denition. The text string will be displayed later
when that denition is called. You may wish to follow it with cr to ush out the text
buffer immediately. Use .( for any printing to be done immediately.


s"
stack: ( [text<">] -- text-str text-len )
generates: b(") len-byte xx-byte xx-byte  xx-byte
Gather the immediately-following string delimited by " . Return the location of the
string text-str text-len.
Since an implementation is only required to provide two temporary buffers, a program
cannot depend on the systems ability to simultaneously maintain more than two
distinct interpreted strings. Compiled strings do not have this limitation, since they are
not stored in the temporary buffers.


==============
bbranch 		(Tokenized by again, repeat, and else)
 -offset for again and repeat
 for repeat, b(>resolve) follows bbranch

 +offset for else
 b(>resolve) follows bbranch
 followed by another b(>resolve) later for then

==============
 b?branch       (Tokenized by until, while, and if)
 -offset for until
 +offset for if and while

==============
 b(>resolve)    (Tokenized by else, then, and repeat)
 Target of forward bbranch or b?branch.

==============
 b(<mark)       (Tokenized by begin)
 target of backward bbranch or b?branch

==============
==============
 then
 Terminate an if statement.
==============

 begin
 Begin a conditional loop.
 Tokenizer equivalent: b(<mark)
==============
 if
 Tokenizer equivalent: b?branch +offset

 while
 Mark first clause of a begin...while...repeat loop.
 Tokenizer equivalent: b?branch +offset
==============
 else
 Tokenizer equivalent: bbranch +offset b(>resolve)
==============
 repeat
 Mark end of a begin...while...repeat loop. Jump to begin.
 Tokenizer equivalent: bbranch -offset b(>resolve)

 again
 End an (infinite) begin...again loop.
 Tokenizer equivalent: bbranch -offset
==============
 until
 End a begin...until loop. Exit loop if flag is nonzero.
 Tokenizer equivalent: b?branch -offset
==============


b(<mark) ... bbranch -offset                                         begin ... again
b(<mark) ... b?branch -offset                                        begin ... until
bbranch +offset ... b(>resolve)                                      ... else ... then
b?branch +offset ... b(>resolve)                                     if ... then
b(<mark) ... b?branch +offset ... bbranch -offset ... b(>resolve)    begin ... while ... repeat


bbranch     ( -- )             Unconditional branch FCode. Followed by offset.
b?branch    ( continue? -- )   Conditional branch FCode. Followed by offset.
b(<mark)    ( -- )             Target of backward branches.
b(>resolve) ( -- )             Target of forward branches.

==========================================================================================
To Do:
Number the structs and include field offset in the name before field size.


==========================================================================================

=end GHOSTCODE
=cut
