/*
 *                     OpenBIOS - free your system! 
 *                            ( detokenizer )
 *                          
 *  decode.c - contains output wrappers for fcode words.  
 *  
 *  This program is part of a free implementation of the IEEE 1275-1994 
 *  Standard for Boot (Initialization Configuration) Firmware.
 *
 *  Copyright (C) 2001-2003 Stefan Reinauer, <stepan@openbios.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>

#include "detok.h"
#include "stream.h"
#include "dictionary.h"
#include "decode.h"

#ifdef USEDISASSEMBLER
#include "Disassembler.h"
#endif

long current_token_pos;
u16 current_token_fcode;
token_t* current_token_record;
char * current_token_name;

int gPass;

int indent, use_tabs=0, verbose=0, decode_all=0, ignore_len=0, mac_rom=0, linenumbers=0;
u32 romstartoffset = 0;

bool offs16;
bool got_start;
bool isSpecialStart;

int linenum;
long fclen;
long filelen;
long gStartPos;
int gLastPos;
int gLastMacRomTokenPos;

bool gLastTokenIsAlias = FALSE;

static void dump_line(int pos)
{
	linenum++;
	if ( gPass )
	{
		if (linenumbers & 1)
			printf("%6d: ",linenum);

		if (linenumbers & 2)
			printf("%06lX: ",pos + romstartoffset);
	}
}


static void dump_buffer_hex( void * buffer, int len, int spaceBytes )
{
	int					index;

	for ( index = 1; index <= len; index++ )
	{
		if (index % spaceBytes == 0)
			printf( "%02X ", *(u8*)buffer++ );
		else
			printf( "%02X", *(u8*)buffer++ );
	}
}


static void make_buffer_text( void * dstBuffer, void * srcBuffer, int len )
{
	u8* dstCurrent = (u8*)dstBuffer;
	
	while ( len-- > 0 )
	{
		u8 c = *(u8*)(srcBuffer)++;
		if ( c < ' ' || c > 0xFE || c == 0x7F )
			c = '.';
		*dstCurrent++ = c;
	}
	*dstCurrent++ = 0;
}


static void dump_hex_one_line( void* buffer, int len )
{
	if ( gPass )
	{
		u8					textBuffer[33];

		if ( len > 32 )
		{
			len = 32;
			fprintf(stderr, "Line %d # Too many hex bytes to dump\n", linenum);
		}

		dump_buffer_hex( buffer, len, 2 );
		make_buffer_text( textBuffer, buffer, len );

		printf( "%*s \"%s\"\n", ((32 - len) * 2 * 5 + 2) / 4, "", textBuffer );
	}
}



static void dumptokenoffset( s32 previous_token_offset, u32 previous_token_address )
{
	if ( gPass )
	{
		char comment[10];
		char instructionText[5];
		dump_buffer_hex( (u8*)&previous_token_offset, 4, 2 );
		make_buffer_text( instructionText, &previous_token_offset, 4 );
		sprintf( comment, "%08lX", previous_token_address );
		printf( "%12s          %-8s %-39s \"%s\"\n", comment, "", "", instructionText );
	}
}

static void dump_hex( long currentPos, long endPos )
{
	u8					buffer[32];
	u8 *				bufferCurrent;
	int					len;
	int					maxlen;

	maxlen = 4 - (currentPos & 3);
	if ( maxlen == 4 )
		maxlen = 32;
	
	if ( !gPass )
	{
		long total = endPos - currentPos;
		if ( total < maxlen )
			maxlen = total;
		total = total - maxlen;
		linenum += ( total + 31 ) / 32 + ( maxlen > 0 );
		set_streampos( endPos );
	}
	else
	{
		set_streampos( currentPos );
		
		while (currentPos < endPos)
		{
			dump_line(currentPos);
			
			len = 0;
			bufferCurrent = buffer;
			while ( (currentPos < endPos) && (len < maxlen) )
			{
				*bufferCurrent++ = get_num8();
				len++;
				currentPos++;
			}

			dump_hex_one_line( buffer, len );
			maxlen = 32;
		}
	}
}


static int myisprint(int c)
{
	return ((c>=' ' && c<='~') || c=='\t');
}


static void pretty_string(s8 *string, u8 len, s8* strbuff)
{
	char* curbuff = strbuff;

#if 0
	u8 c;
	bool qopen=TRUE;
	
	curbuff += sprintf(curbuff, "\"");

	while ((c=(*(string++)))) {
		if (isprint(c)) {
			if (!qopen) {
				curbuff += sprintf(curbuff, "\"");
				qopen=TRUE;
			}
			curbuff += sprintf(curbuff, "%c",c);
		} else {
			if (qopen) {
				curbuff += sprintf(curbuff, "\"");
				qopen=FALSE;
			}
			curbuff += sprintf(curbuff, " 0x%02x",c);
		}
	}
	if (qopen)
		curbuff += sprintf(curbuff, "\"");
#else
	u8 c;
	bool qopen=TRUE;

	u8 * cptr;
	int numleft;
	bool hexonly = FALSE;
	int ndx = 0;
	
	cptr = (u8 *)string;
	for ( numleft = len; numleft > 0; numleft-- )
	{
		c=(*(cptr++));
		if ( ( c == 0 && len >= 252 ) || ( c != 0 && myisprint(c) == FALSE ) )
		{
			hexonly = TRUE;
			break;
		}
	}

	curbuff += sprintf(curbuff, "\" ");

	cptr = (u8 *)string;
	for ( numleft = len; numleft > 0; numleft-- )
	{
		c=(*(cptr++));
		if (hexonly == FALSE && c == '"') {
			if (!qopen) {
				curbuff += sprintf(curbuff, ")\"\"");
				qopen=TRUE;
			}
			else
				curbuff += sprintf(curbuff, "\"\"");
		} else if (hexonly == FALSE && myisprint(c)) {
			if (!qopen) {
				curbuff += sprintf(curbuff, ")%c",c);
				qopen=TRUE;
			}
			else
				curbuff += sprintf(curbuff, "%c",c);
		} else {
			if (qopen) {
				curbuff += sprintf(curbuff, "\"(%02X",c);
				qopen=FALSE;
			}
			else
				if ( ndx % 2 )
					curbuff += sprintf(curbuff, "%02X",c);
				else
					curbuff += sprintf(curbuff, " %02X",c);
		}
		ndx++;
	}
	if (!qopen) {
		curbuff += sprintf(curbuff, ")");
		qopen=TRUE;
	}

	if ( hexonly )
	{
		u8 buffer[256];
		make_buffer_text( buffer, string, len );
		curbuff += sprintf(curbuff,  "\" \\ \"%s\" ", buffer );
	}
	else
		curbuff += sprintf(curbuff, "\" ");
#endif
}

static void decode_indent(void) 
{
	if ( gPass )
	{
		int i;

		if (indent<0) {
			fprintf(stderr, "Line %d # Indentation error\n", linenum);
			indent=0;
		}

		if (use_tabs)
			for (i=0; i<indent; i++)
				printf ("\t");
		else
			for (i=0; i<indent; i++)
				printf ("    ");
	}
}


static void mark_last_streampos(void)
{
	gLastPos = get_streampos();
}


#ifdef USEDISASSEMBLER
DisassemblerOptions	options = Disassemble_PowerPC32 | Disassemble_Extended | Disassemble_DecField | Disassemble_DecOffset | Disassemble_DecPCRel | Disassemble_CRBits | Disassemble_BranchBO | Disassemble_TrapTO;


static char * EmptyDisassemblerLookups(
				void*							refCon,
				const unsigned long *			cia, 
				const DisassemblerLookupType	lookupType,
				const DisLookupValue 	 		thingToReplace)
{
	#pragma unused( refCon, cia, thingToReplace )
	switch ( lookupType )
	{
		case Disassembler_Lookup_SPRegister:
			#warning "I don't think this works with HID1; please test instructions 7E91FAA6 and 7E91FBA6"
			return NULL;	/* the default lookup will detect unknown special purpose registers */
		default:
			return "_";
	}
}


u32					instruction;
u32					gSpecialPurposeRegister;
token_t*			gDissasembledToken = NULL;

static char * MyDisassemblerLookups(
				void*							refCon,
				const unsigned long *			cia, 
				const DisassemblerLookupType	lookupType,
				const DisLookupValue 	 		thingToReplace)
{
	#pragma unused( refCon, cia )
	static char substitute_string[100];
	
	gDissasembledToken = NULL;

	switch ( lookupType )
	{
		case Disassembler_Lookup_GPRegister:
		{
			if ( thingToReplace.gpr == 2 )
				return "r2";
			if ( thingToReplace.gpr == 1 )
				return "r1";
			break;
		}
		
		case Disassembler_Lookup_RelAddress:
		{
			long tokenOffset;

			if ( thingToReplace.relAddress - romstartoffset >= gLastPos )
				break;

			gDissasembledToken = find_token_by_pos( thingToReplace.relAddress - romstartoffset, &tokenOffset );
		
			if ( gDissasembledToken == NULL )
				break;

			if ( tokenOffset == 0 )
				return gDissasembledToken->name;
				
			sprintf( substitute_string, "%s%+ld", gDissasembledToken->name, tokenOffset );
			{
				gDissasembledToken = NULL;
				return substitute_string;
			}

			break;
		}
		
		case Disassembler_Lookup_SImmediate:
		{
			if ( (instruction & 0xFC1F0000) == 0x3C000000 ) /* lis */
			{
				if ( options & Disassemble_DollarHex )
					sprintf( substitute_string, "$%04X", thingToReplace.si & 0x0FFFF );
				else
					sprintf( substitute_string, "0x%04X", thingToReplace.si & 0x0FFFF );
				return substitute_string;
			}
			else
			{
				u32 num = abs( thingToReplace.si );
				
				if
					(
							( num < 10 )
						||	( num % 10 == 0 && num <= 100 )
						||	( num % 100 == 0 && num <= 10000 )
						||	( num % 1000 == 0 )
					)
				{
					sprintf( substitute_string, "%d", thingToReplace.si );
					return substitute_string;
				}
			}
			break;
		}

		case Disassembler_Lookup_UImmediate:
		{
				u32 num = thingToReplace.ui;
				if
					(
							( num < 10 )
						||	( num % 10 == 0 && num <= 100 )
						||	( num % 100 == 0 && num <= 10000 )
						||	( num % 1000 == 0 )
					)
				{
					sprintf( substitute_string, "%u", thingToReplace.ui );
					return substitute_string;
				}
				break;
		}
		
		case Disassembler_Lookup_SPRegister:
		{
			gSpecialPurposeRegister = thingToReplace.spr;

			switch( thingToReplace.spr )
			{
				/* G3 (MPC750) registers - don’t exist on the G5 */
				case 1017:
					return "L2CR"; /* L2 Cache Control Register (L2CR) */
				case 1019:
					return "ICTC"; /* Instruction Cache Throttling Control Register (ICTC) */
				
				
				default:
					if ( mac_rom & 4 )
						/* G5 (PPC 970fx) registers */
						switch( thingToReplace.spr )
						{
							case 276:
								return "SCOMC"; /* Scan Communications Register (SCOMC) */
							case 277:
								return "SCOMD"; /* Scan Communications Register (SCOMD) */
							case 311:
								return "HIOR"; /* Hardware Interrupt Offset Register (HIOR) */
							case 1012:
								return "HID4"; /* Hardware Implementation Register (HID4) */
							case 1014:
								return "HID5"; /* Hardware Implementation Registers (HID5) */
							case 1015:
								return "DABRX"; /* Data Address Breakpoint Register (DABRX) */
						}
			}
			break;
		}
	}
	
	return NULL;
}


static void dumptokenline( u32 tokenAddress, u32 theNum, u32 currentPos )
{
	if ( gPass )
	{
		char				operand[100];
		char				instructionText[5];
		char				comment[10];

		if ( tokenAddress == 0 )
		{
			strcpy( operand, "0" );
		}
		else
		{
			s32					tokenOffset;
			token_t*			theToken;
			s32					offsetFromTokenStart;

			tokenOffset = tokenAddress - romstartoffset;
			theToken = find_token_by_pos( tokenOffset, &offsetFromTokenStart );
		
			if ( theToken == NULL )
				sprintf( operand, "$%+ld", tokenOffset - currentPos );
			if ( offsetFromTokenStart )
				sprintf( operand, "%s%+ld", theToken->name, offsetFromTokenStart );
			else
				strcpy( operand, theToken->name );
		}

		dump_line( currentPos );
		dump_buffer_hex( &theNum, 4, 2 );
		make_buffer_text( instructionText, &theNum, 4 );

		sprintf( comment, "%08lX", tokenAddress );
		printf( "%12s          %-8s %-39s \"%s\"\n", comment, "dc.l", operand, instructionText );
	}
	else
		linenum++;
}


static void dumpliteralline( s32 theNum, u32 currentPos )
{
	if ( gPass )
	{
		char				operand[20];
		char				instructionText[5];

		u32 absliteral;
		absliteral = labs( theNum );
		if
			(
					( absliteral < 10 )
				||	( absliteral % 10 == 0 && absliteral <= 100 )
				||	( absliteral % 100 == 0 && absliteral <= 10000 )
				||	( absliteral % 1000 == 0 )
			)
			sprintf( operand, "%ld", theNum );
#ifdef USEDISASSEMBLER
		else if ( options & Disassemble_DollarHex )
			sprintf( operand, "$%lX", theNum );
#endif
		else
			sprintf( operand, "0x%lX", theNum );

		dump_line( currentPos );
		dump_buffer_hex( &theNum, 4, 2 );
		make_buffer_text( instructionText, &theNum, 4 );
		printf( "%12s          %-8s %-39s \"%s\"\n", "", "dc.l", operand, instructionText );
	}
	else
		linenum++;
}


static void dumpoffsetline( s32 theNum, u32 currentPos )
{
	if ( gPass )
	{
		char				operand[20];
		char				instructionText[5];

		if ( theNum < 100 )
			sprintf( operand, "%ld", theNum );
#ifdef USEDISASSEMBLER
		else if ( options & Disassemble_DollarHex )
			sprintf( operand, "$%04lX", theNum );
#endif
		else
			sprintf( operand, "0x%04lX", theNum );
										
		dump_line( currentPos );
		dump_buffer_hex( &theNum, 4, 2 );
		make_buffer_text( instructionText, &theNum, 4 );
		printf( "%12s          %-8s %-39s \"%s\"\n", "", "dc.l", operand, instructionText );
	}
	else
		linenum++;
}
#endif /* USEDISASSEMBLER */


static void decode_lines(void)
{
#ifdef USEDISASSEMBLER
	DisassemblerStatus	disasmStat;
	char				mnemonic[100];
	char				operand[100];
	char				comment[200];
	char*				comment2;
	char				instructionText[5];
#endif

	long				endPos = get_streampos();
	long				currentPos = gLastPos;


	if ( currentPos > endPos )
		if ( !gPass )
			fprintf( stderr, "Line %d # currentPos %lX > endPos %lX)\n", linenum + 1, currentPos, endPos );

	set_streampos( currentPos );

#ifdef USEDISASSEMBLER

	if (mac_rom)
	{
		long adjust = romstartoffset - (long)&instruction;
		long instructionEnd = endPos & -4; /* 3->0  4->4 */

		while (currentPos < endPos)
		{
/* loop through bad PPC instructions */
			
			long firstHexPos = currentPos;
			currentPos = (currentPos + 4-1) & -4; /* 0->0  1->4  4->4 */

			if ( !gLastTokenIsAlias )
			{
				while (currentPos < instructionEnd)
				{
					instruction = get_num32();

					if ( instruction != 0xDEADBEEF )
					{
						disasmStat = Disassembler_OK ^ ppcDisassembler(
														&instruction,
														adjust + currentPos,
														options,
														mnemonic,
														operand,
														comment,
														NULL, /* refCon */
														EmptyDisassemblerLookups  /* DisassemblerLookups lookupRoutine */
													);

						if ( instruction == 0x7C2002A6 ) /* this instruction "mfspr r1,MQ ; 0" has the Disassembler_InvField bit set because it is 601 specific */
							disasmStat &= ~Disassembler_InvField;
						if ( ( disasmStat & ( Disassembler_OK | Disassembler_InvRsvBits | Disassembler_InvField /*| Disassembler_Privileged*/ ) ) == 0 )
						{
							set_streampos( currentPos );
							break; /* got a good instruction */
						}
					}

					currentPos += 4;
				} /* while */
			}
			
			if ( currentPos == instructionEnd )
				currentPos = endPos;

/* dump bad PPC instructions */
			dump_hex( firstHexPos, currentPos );

/* loop through good PPC instructions */
			while (currentPos < instructionEnd)
			{
				instruction = get_num32();

				if ( gLastTokenIsAlias )
				{
					u32 tokenAddress = instruction + currentPos + romstartoffset;
					dumptokenline( tokenAddress, instruction, currentPos );
					currentPos += 4;
					gLastTokenIsAlias = FALSE;
				}
				else
				{
					if ( instruction != 0xDEADBEEF )
					{
						disasmStat = Disassembler_OK ^ ppcDisassembler(
														&instruction,
														adjust + currentPos,
														options,
														mnemonic,
														operand,
														comment,
														NULL, /* refCon */
														MyDisassemblerLookups  /* DisassemblerLookups lookupRoutine */
													);

						if ( instruction == 0x7C2002A6 ) /* this instruction "mfspr r1,MQ ; 0" has the Disassembler_InvField bit set because it is 601 specific */
						{
							disasmStat &= ~Disassembler_InvField;
							fprintf( stderr, "Line %d # Assembly instruction has invalid field (MQ is a PPC601 specific special purpose register)\n", linenum + 1 );
						}
						
						if ( ( disasmStat & Disassembler_InvSprMaybe ) != 0 )
						{
							if ( !gPass )
								fprintf( stderr, "Line %d # Possibly invalid special purpose register %ld\n", linenum + 1, gSpecialPurposeRegister );
						}


					}
					else
						disasmStat = (DisassemblerStatus)-1;

					if ( ( disasmStat & ( Disassembler_OK | Disassembler_InvRsvBits | Disassembler_InvField /*| Disassembler_Privileged*/ ) ) != 0 )
					{
						set_streampos( currentPos );
	/* got bad instruction so break */
						break;
					}

					if ( gPass )
					{
						dump_line( currentPos );
						dump_buffer_hex( &instruction, 4, 2 );
						make_buffer_text( instructionText, &instruction, 4 );

						comment2 = "";

						if ( instruction == 0x7C0006AC ) /* eieio */
							comment[0] = 0;
						else if ( strstr( comment, " = " ) )
							comment[0] = 0;
						else if ( strlen( comment ) == 12 )
						{
							if ( memcmp( "; 0x00", comment, 6 ) == 0 )
							{
								comment[0] = 0;
								comment2 = comment + 6;
							}
							else if ( memcmp( "; 0x", comment, 4 ) == 0 )
							{
								comment[0] = 0;
								comment2 = comment + 4;
							}
						}

						if ( comment[0] )
							printf( "%12s          %-8s %-22s %-16s \"%s\"\n", comment2, mnemonic, operand, comment, instructionText );
						else
							printf( "%12s          %-8s %-39s \"%s\"\n", comment2, mnemonic, operand, instructionText );
					}
					else
						linenum++;

					currentPos += 4;

					if ( gDissasembledToken != NULL )
					{
						if ( currentPos < instructionEnd && !strcmp( mnemonic, "bl" ) )
						{
							u32 theNum;
							u32 tokenAddress;

							if ( !strcmp( gDissasembledToken->name, "{'}" ) )
							{
								theNum = get_num32(); /* bl tokenname instruction */
								tokenAddress = (((s32)theNum & 0x03FFFFFC) << 6) / 64 + currentPos + romstartoffset;
								dumptokenline( tokenAddress, theNum, currentPos );
								currentPos += 4;
							}
							else
							if ( !strcmp( gDissasembledToken->name, "b<\">" ) )
							{
								bool isGoodString;
								u8* string;
								u8 len;

								string = get_string2( &len, &isGoodString );
								if ( string && isGoodString )
								{
									if ( gPass )
									{
										s8 strbuff[2000];
									
										pretty_string(string, len, strbuff);

										dump_line( currentPos );
										printf( "%02X ...    %12s          %-8s %s\n", len, "", "dc.b", strbuff );
									}
									else
										linenum++;

									currentPos += ((len + 1 + 3) & -4);
									if ( currentPos > instructionEnd )
									{
										currentPos = instructionEnd;
										if ( !gPass )
											printf( stderr, "Line %d # string litteral extends too far\n", linenum );
									}
								}
								free(string);
								set_streampos( currentPos );
							}
							else
							if ( !strcmp( gDissasembledToken->name, "b<lit>" ) || !strcmp( gDissasembledToken->name, "(val)" ) )
							{
								theNum = get_num32();
								dumpliteralline( theNum, currentPos );
								currentPos += 4;
							}
							else
							if ( !strcmp( gDissasembledToken->name, "(i-val)" ) || !strcmp( gDissasembledToken->name, "(i-var)" ) || !strcmp( gDissasembledToken->name, "(field)" ) )
							{
								theNum = get_num32(); /* offset in instance or struct */
								dumpoffsetline( theNum, currentPos );
								currentPos += 4;
							}
							else
							if ( !strcmp( gDissasembledToken->name, "(v-val)" ) )
							{
								theNum = get_num32();
								dumpoffsetline( theNum, currentPos );
								currentPos += 4;
								theNum = get_num32();
								dumpliteralline( theNum, currentPos );
								currentPos += 4;
								theNum = get_num32();
								dumptokenline( theNum, theNum, currentPos );
								currentPos += 4;
							}
							else
							if ( !strcmp( gDissasembledToken->name, "(i-defer)" ) )
							{
								theNum = get_num32();
								dumpoffsetline( theNum, currentPos );
								currentPos += 4;
								theNum = get_num32();
								dumptokenline( theNum, theNum, currentPos );
								currentPos += 4;
							}
							else
							if ( !strcmp( gDissasembledToken->name, "(v-defer)" ) )
							{
								theNum = get_num32();
								dumpoffsetline( theNum, currentPos );
								currentPos += 4;
								theNum = get_num32();
								dumptokenline( theNum, theNum, currentPos );
								currentPos += 4;
								theNum = get_num32();
								dumptokenline( theNum, theNum, currentPos ); /* storage */
								currentPos += 4;
							}
							else
							if ( !strcmp( gDissasembledToken->name, "b<to>" ) || !strcmp( gDissasembledToken->name, "b<'>" ) || !strcmp( gDissasembledToken->name, "(i-to)" ) || !strcmp( gDissasembledToken->name, "(v-to)" ) || !strcmp( gDissasembledToken->name, "(defer)" ) )
							{
								theNum = get_num32();
								tokenAddress = theNum;
								dumptokenline( tokenAddress, theNum, currentPos );
								currentPos += 4;
							}
							else
							if ( !strcmp( gDissasembledToken->name, "b<to>1" ) )
							{
								theNum = get_num32();
								tokenAddress = theNum + currentPos + romstartoffset;
								dumptokenline( tokenAddress, theNum, currentPos );
								currentPos += 4;
							}
						} /* if ( currentPos < instructionEnd && !strcmp( mnemonic, "bl" ) ) */
					} /* if gDissasembledToken */
				} /* if gLastTokenIsAlias else */
			} /* while good instructions */
		} /* while */
	} /* if mac_rom */
#endif /* USEDISASSEMBLER */

	dump_hex( currentPos, endPos );

	dump_line(endPos);
}

static void output_token(void)
{
	if ( gPass )
	{
		char * tname = lookup_token( fcode );

		if (tname == unnamed)
			printf( "unnamed_fcode_%03x ", fcode );
		else
			printf ("%s ", tname);

		/* The fcode number is interesting if
		 *  (a) detok is in verbose mode
		 *  (b) the token has no name.
		 */
			
		if (verbose)
			printf("\\ [0x%03x] ", fcode);

		if (tname == fcerror && indent > 20)
			indent=0;
	}
}

static s16 decode_offset(void)
{
	s16 offs;

	long offsetaddr = get_streampos();

	offs=get_offset();

	if (offs == 0)
	{
		if ( gPass )
			fprintf(stderr, "Line %d # Offset 0, switching to 16 bit offsets\n", linenum);
		set_streampos( offsetaddr );
		offs16 = TRUE;
		offs=get_offset();
	}

	if ( gPass )
	{
		output_token();
		printf("0x%x\n",offs);
	}
	return offs;
}

static void decode_default(void)
{
	if ( gPass )
	{
		output_token();
		printf ("\n");
	}
}

static void new_token(void)
{
	u16 token;
	output_token();
	token=get_token();
	if ( gPass )
		printf("0x%03x\n",token);
	add_token(token, unnamed);
}


static bool good_tokenname(char* tokenname, u8 len)
{
	size_t pos;
	return (pos=strspn(tokenname, "!\"#$%&'()*+,-./0123456789:;<=>?@[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~")) && (tokenname[pos]==0) && (pos==len);
}

static void named_token(bool isExternal)
{
	#pragma unused(isExternal)
	u16 token;
	char* string;
	int pos;
	u8 len;

	output_token();

	pos=get_streampos();
	string=(char*)get_string(&len);
	token=get_token();

	if (good_tokenname(string, len))
	{
		if ( gPass )
			printf("%s 0x%03x\n", string, token);
		add_token(token,string);
	}
	else
	{
		char* p;
		int numleft;

		set_streampos(pos);

		if ( gPass )
		{
			numleft = len;
			for (p=string; numleft; numleft--,p++)
				if (*p==13 || *p==10 || *p==0)
					*p = ' ';

			printf("%s 0x%03x \\ bad name, ignored\n", string, token);
			fprintf(stderr, "Line %d # Bad token name: \"%s\"\n", linenum, string);
		}

		free(string);
	}
}


/* Old World G3, 8600 flags also used by G5 */
										
										/* flag names defined in G5 Open Firmware */
										
#define tokenFlagDefined		0x80	/* fdefd		- token is completely defined and ready to be called */
#define tokenFlagImmediate		0x40	/* fimm			- tokens like instance, vectored, b(lit), b('), b("), b(<mark), b(<resolve), b(>mark), b(>resolve), bbranch, lbbranch, b(do), b(case), external-token, if, case, then, of, while, repeat, do, */
#define tokenFlagNoHeader 		0x20	/* fnohdr		- token created with new-token (instead of external-token or named-token) while headerless (instead of external or headers) */
#define tokenFlagAlias			0x10	/* falias		- token created with alias */
#define tokenFlagInstance		0x08	/* finstance	- token created with instance */
/* G5 only flags */
#define tokenFlagVisible		0x04	/* fvisible		- token created with external-token (instead of named-token or new-token) while external (instead of headers or headerless) */
#define tokenFlagInvisible		0x02	/* finvisible	- token created with named-token (instead of external-token or new-token) while headers (instead of external or headerless) */
#define tokenFlagVectored		0x01	/* fvectored	- token created with vectored */


static void decode_rom_token()
{
	u16 definedFCodeNumber;
	s8* stringFromFile;
	s8* stringForTokenName;
	s8* stringToDelete = NULL;
	u8 len;
	u8 mac_rom_fcode_flags;
	bool good_token_name;
	long defined_token_pos, after_defined_token_pos, after_string_pos, token_code_pos;
	token_t* theToken;

	token_t* aliasToken = NULL;
	long alias_token_offset;
	long alias_token_pos;

	int numZeros;
	bool isGoodString;


/*
value
  3767: 0045F9: b(value) [0x0b8] 0x4e5
  3768: 0045FC: 7E48 02A6 						mflr     r18                                     "~H.¶"
  3769: 004600: 4BFF C8C5 						bl       $-0x373C               ; 0x00000EC4     "K.»≈"
  3770: 004604: 0000 0000                                                                        "........"

   808: 000EC1: b(code) [0x0bf] 0x432
   809: 000EC4: 7E49 03A6 						mtctr    r18                                     "~I.¶"
   810: 000EC8: 7C68 02A6 						mflr     r3                                      "|h.¶"
   811: 000ECC: 969F FFFC 						stwu     r20,-0x0004(r31)                        "ñü.¸"
   812: 000ED0: 8283 0000 						lwz      r20,0x0000(r3)                          "ÇÉ.."
   813: 000ED4: 4E80 0420 						bctr                                             "NÄ. "

buffer
  3788: 004671: b(buffer:) [0x0bd] 0x4ea
  3789: 004674: 7E48 02A6 						mflr     r18                                     "~H.¶"
  3790: 004678: 4BFF C84D 						bl       $-0x37B4               ; 0x00000EC4     "K.»M"
  3791: 00467C: 0000 0100                                                                        "...."

   808: 000EC1: b(code) [0x0bf] 0x432
   809: 000EC4: 7E49 03A6 						mtctr    r18                                     "~I.¶"
   810: 000EC8: 7C68 02A6 						mflr     r3                                      "|h.¶"
   811: 000ECC: 969F FFFC 						stwu     r20,-0x0004(r31)                        "ñü.¸"
   812: 000ED0: 8283 0000 						lwz      r20,0x0000(r3)                          "ÇÉ.."
   813: 000ED4: 4E80 0420 						bctr                                             "NÄ. "

variable
  3763: 0045E1: b(variable) [0x0b9] 0x4e4
  3764: 0045E4: 7E48 02A6 						mflr     r18                                     "~H.¶"
  3765: 0045E8: 4BFF C9BD 						bl       $-0x3644               ; 0x00000FA4     "K.…Ω"
  3766: 0045EC: 0000 0000                                                                        "...."


   861: 000FA1: b(code) [0x0bf] 0x437
   862: 000FA4: 7E49 03A6 						mtctr    r18                                     "~I.¶"
   863: 000FA8: 969F FFFC 						stwu     r20,-0x0004(r31)                        "ñü.¸"
   864: 000FAC: 7E88 02A6 						mflr     r20                                     "~à.¶"
   865: 000FB0: 4E80 0420 						bctr                                             "NÄ. "

constant
  3804: 0046D1: b(constant) [0x0ba] 0x4ee
  3805: 0046D4: 969F FFFC 						stwu     r20,-0x0004(r31)                        "ñü.¸"
  3806: 0046D8: 3A80 0018 						li       r20,0x0018                              ":Ä.."
  3807: 0046DC: 4E80 0020 						blr                                              "NÄ. "

  3864: 0047F1: b(constant) [0x0ba] 0x4fa
  3865: 0047F4: 969F FFFC 						stwu     r20,-0x0004(r31)                        "ñü.¸"
  3866: 0047F8: 3E80 0000 						lis      r20,0x0000                              ">Ä.."
  3867: 0047FC: 6294 FFFF 						ori      r20,r20,0xFFFF                          "bî.."
  3868: 004800: 4E80 0020 						blr                                              "NÄ. "

  3881: 004839: b(constant) [0x0ba] 0x4fd
  3882: 00483C: 969F FFFC 						stwu     r20,-0x0004(r31)                        "ñü.¸"
  3883: 004840: 3E80 4800 						lis      r20,0x4800                              ">ÄH."
  3884: 004844: 4E80 0020 						blr                                              "NÄ. "

field
  3898: 004899: b(field) [0x0be] 0x502
  3899: 00489C: 3A94 0000 						addi     r20,r20,0x0000                          ":î.."
  3900: 0048A0: 4E80 0020 						blr                                              "NÄ. "
  3901: 0048A4: FFFF FFF0                                                                        "..."

alias
  3892: 004869: b(constant) [0x0ba] 0x4ff
  3893: 00486C: FFFF FF28                                                                        "...("
  3894: 004879: b(constant) [0x0ba] 0x500
  3895: 00487C: FFFF FF48                                                                        "...H"
  3896: 004889: b(constant) [0x0ba] 0x501
  3897: 00488C: FFFF FF20                                                                        "... "

set-token
 14041: 00E1FD: set-token [0x0db] 0xe19 %r0
 14043: 00E208: 967E FFFC                       stwu     r19,-4(r30)                             "ñ~.¸"
 14044: 00E20C: 7E68 02A6                       mflr     r19                                     "~h.¶"
 14045: 00E210: 4BFF FEE9       00E0F8          bl       mac_rom_colon_70C+64                    "K.˛È"

*/



/* Check token file pos */

	defined_token_pos = get_streampos();
	if ( (defined_token_pos & 3) != 2 )
		return;

/* Check token FCode number */

	definedFCodeNumber = get_num16();
	if
		(
			(
				(mac_rom & 4) == 0
				&&
				definedFCodeNumber > 0xFFF
			)
#if 0
			||
			(
				(mac_rom & 4) != 0
				&&
				definedFCodeNumber >  0x3FFF /* the G5 has FCode's greater than 0xFFF */
				&&
				definedFCodeNumber < 0xC000 /* the G5 uses FCode 0xFFFF for user defined words */
			)
#endif
			||
			(
				(definedFCodeNumber < 0x00F)
				&&
				(definedFCodeNumber != 0x000) /* defined by b(:) in mac rom */
			)
		)
	{
		set_streampos( defined_token_pos );
		return;
	}

/* Check token flags */

	set_streampos( current_token_pos - 1 );
	mac_rom_fcode_flags = get_num8();

	if	(
			( ( ( /* tokenFlagUnusedFlags | */ tokenFlagDefined) & mac_rom_fcode_flags ) != tokenFlagDefined )
		)
	{
		set_streampos( defined_token_pos );
		return;
	}

/* Get token name */

	after_defined_token_pos = defined_token_pos + 2;

	set_streampos( after_defined_token_pos );
	stringFromFile = get_string2( &len, &isGoodString );

	good_token_name = stringFromFile && isGoodString && good_tokenname( stringFromFile, len );
	if ( good_token_name != ( ( mac_rom_fcode_flags & tokenFlagNoHeader ) == 0 ) )
	{
		set_streampos( defined_token_pos );
		return;
	}

/* Find offset of token code */

	if ( good_token_name )
	{
		after_string_pos = get_streampos();
		numZeros=0;
		while (get_num8()==0 && numZeros<256)
			numZeros++;
		token_code_pos = get_streampos() - 1; /* after the last zero */
	}
	else
		token_code_pos = after_defined_token_pos;

/* Check alias token */

	if ( mac_rom_fcode_flags & tokenFlagAlias )
	{
		set_streampos( token_code_pos );
		alias_token_pos = token_code_pos + get_num32();
		aliasToken = find_token_by_pos( alias_token_pos, &alias_token_offset );
		if ( !aliasToken )
		{
			set_streampos( defined_token_pos );
			return;
		}
	}

/* Check last token offset and dump everything since then */
	{
		long caclulated_previous_rom_token_pos;
		long previous_token_offset;

		set_streampos(current_token_pos - 5);
		previous_token_offset = get_num32();

		if ( previous_token_offset == 0 )
		{
			set_streampos(current_token_pos - 5);
			decode_lines();
			dump_hex_one_line( &previous_token_offset, 4 );
			if ( gPass )
				fprintf(stderr, "Line %d # Offset %06lX; Previous token offset is 0\n", linenum, current_token_pos - 5 + romstartoffset);
		}
		else
		{
			caclulated_previous_rom_token_pos = previous_token_offset + current_token_pos - 5;
		
			if ( caclulated_previous_rom_token_pos == gLastMacRomTokenPos )
			{
				set_streampos(current_token_pos - 5);
				decode_lines();
				dumptokenoffset( previous_token_offset, caclulated_previous_rom_token_pos + romstartoffset );
			}
			else if ( previous_token_offset < 0 && caclulated_previous_rom_token_pos >= 0 )
			{
				long insidetokenoffset;
				token_t * backtoken = find_token_by_offsetpos( caclulated_previous_rom_token_pos, &insidetokenoffset);
				
				if ( backtoken != NULL && insidetokenoffset == 0 )
				{
					set_streampos(current_token_pos - 5);
					decode_lines();
					dumptokenoffset( previous_token_offset, caclulated_previous_rom_token_pos + romstartoffset );
					if ( !gPass )
						fprintf(stderr, "Line %d # Offset %06lX; Token offset points to %s at %06lX\n", linenum, current_token_pos - 5 + romstartoffset, backtoken->name, backtoken->execution_pos + romstartoffset );
				}
				else
				{
					if ( !gPass )
						fprintf(stderr, "Line %d # Offset %06lX; Previous token offset %06lX -> %06lX doesn’t have a token\n", linenum, current_token_pos - 5 + romstartoffset, previous_token_offset + romstartoffset, caclulated_previous_rom_token_pos + romstartoffset );
					set_streampos( defined_token_pos );
					return;
				}
			}
			else
			{
				/* offset is bad - don’t report anything */
				set_streampos( defined_token_pos );
				return;
			}

			if ( previous_token_offset & 7 )
				if ( !gPass )
					fprintf(stderr, "Line %d # Offset %06lX; Previous token offset should be a multiple of 8\n", linenum, current_token_pos - 5 + romstartoffset);
		}
	}

/* All good; mark it */

	gLastMacRomTokenPos = current_token_pos - 5;

/* Dump token flags */

	dump_line(current_token_pos - 1);
	dump_hex_one_line( &mac_rom_fcode_flags, 1 );

/* Dump token type */

	dump_line(current_token_pos);
	output_token();

/* Check alias token */

	if ( aliasToken && alias_token_offset == 0 )
	{
		gLastTokenIsAlias = TRUE;
		set_streampos( alias_token_pos ); /* for token naming purposes; if it doesn't have a name */
	}
	else
		set_streampos( token_code_pos );

/* dump token name */

	if ( good_token_name )
	{
		if ( !gPass )
		{
			int width = numZeros + len + 1;
			if ( ( width != 8 && width != 16 && width != 24 && width != 32 ) || numZeros > 7 )
				fprintf( stderr, "Line %d # zeros:%d width:%d\n", linenum, numZeros, width );
			stringForTokenName = stringFromFile;
		}
	}
	else
	{
		{
			/* the name from the file was bad so clean it up if it has any length */
			int numleft;
			char* p;

			numleft = len;
			for (p=stringFromFile; numleft; numleft--,p++)
				if (*p==13 || *p==10 || *p==0)
					*p = ' ';
		}

		if ( !gPass )
		{
			char *token_class;
			int new_token_len;
			char new_token_name[50];
			char* oldname;
			u32 instructions[5];
			long number_for_name;
			bool number_is_good;

			if ( ( mac_rom_fcode_flags & tokenFlagNoHeader ) == 0 )
				fprintf(stderr, "Line %d # Bad %s token name (len:%d): \"%s\"\n", linenum, current_token_name, len, stringFromFile);

			theToken = find_token(definedFCodeNumber);
			oldname = get_token_name(theToken);

			if (theToken == NULL || oldname == unnamed || definedFCodeNumber >= 0x600 )
			{
				/* make up a name based on token type and contents if the token was not defined before or was unnamed or is a user defined fcode */

				switch (current_token_fcode)
				{
					case 0x0bf: /* b(code) */
						token_class = "code";
						goto DoNoNumber;
						
					case 0x0b7: /* b(:) */
						token_class = "colon";
						goto DoNoNumber;
						
					case 0x0bb: /* b(create) */
						token_class = "create";
						goto DoNoNumber;
						
					case 0x0bc: /* b(defer) */
						token_class = "defer";
						goto DoNoNumber;

					case 0x0db: /* set-token */
						token_class = "settoken";
		DoNoNumber:
						new_token_len = sprintf( new_token_name, "mac_rom_%s_%03X", token_class, definedFCodeNumber );
						break;

					case 0x0b8: /* b(value) */
						token_class = "value";
						goto DoInlineNumber;
						
					case 0x0b9: /* b(variable) */
						token_class = "variable";
						goto DoInlineNumber;

					case 0x0bd: /* b(buffer:) */
					{
						token_class = "buffer";
		DoInlineNumber:
						instructions[0] = get_num32();
		DoInlineNumber2:
						instructions[1] = get_num32();
						instructions[2] = get_num32();

						number_for_name = instructions[2];
						number_is_good = instructions[0] == 0x7E4802A6 && (instructions[1] & 0xFF000000) == 0x4B000000; /* FE000000 0x4A000000 */

		DoCheckNumber:
						if (number_is_good)
						{
							new_token_len = sprintf( new_token_name, "mac_rom_%s_%03X_%lX", token_class, definedFCodeNumber, number_for_name );
						}
						else
						{
							new_token_len = sprintf( new_token_name, "mac_rom_%s_%03X", token_class, definedFCodeNumber );
							if (!number_is_good)
								fprintf(stderr, "Line %d # Bad %s definition for \"%s\"\n", linenum, current_token_name, new_token_name);
						}
						break;
					}
					case 0x0ba: /* b(constant) */
					{
						bool gotNum = FALSE;
						int index;
						number_is_good = FALSE;
						
						token_class = "constant";
						
						instructions[0] = get_num32();
						instructions[1] = get_num32();
						instructions[2] = get_num32();
						instructions[3] = get_num32();

						if ( instructions[0] == 0x969FFFFC )
						{
							for ( index = 1; index < 4; index++ )
							{
								if ( instructions[index] == 0x4E800020 ) /* blr */
								{
									number_is_good = gotNum;
									break;
								}

								switch ( instructions[index] & 0xFFFF0000 )
								{
									case 0x3A800000: /* li */
										number_for_name = (long)(short)( instructions[index] & 0x0000FFFF );
										gotNum = TRUE;
										break;
									case 0x3E800000: /* lis */
										number_for_name = instructions[index] << 16;
										gotNum = TRUE;
										break;
									case 0x62940000: /* ori */
										number_for_name |= instructions[index] & 0x0000FFFF;
										break;
									default:
										goto DoCheckNumber;
								}
							}
						}
						goto DoCheckNumber;
					}

					case 0x0be: /* b(field) */
					{
						token_class = "field";

						instructions[0] = get_num32();
						if ( instructions[0] == 0x7E4802A6 )
							goto DoInlineNumber2;

						instructions[1] = get_num32();

						number_for_name = (long)(short)(instructions[0] & 0x0FFFF);
						number_is_good = (instructions[0] & 0xFFFF0000) == 0x3A940000 && instructions[1] == 0x4E800020;
						goto DoCheckNumber;
					}
					
				} /* switch */

				/* use new name */
				stringForTokenName = malloc( new_token_len + 1 );
				if (stringForTokenName)
					strcpy( stringForTokenName, new_token_name );
				else
				{
					fprintf( stderr, "Line %d # Not enough memory\n", linenum );
					exit(-ENOMEM);
				}
			}
			else
			{
				/* force use the old name if the token was defined and not unnamed and was not a user defined fcode */
				stringForTokenName = NULL;
			}

			stringToDelete = stringFromFile;
		} /* if ( !gPass ) */
	}

/* define token  */
	if ( !gPass )
	{
		theToken = add_token( definedFCodeNumber, stringForTokenName );
		theToken->execution_pos = token_code_pos;
		theToken->offset_pos = gLastMacRomTokenPos;
		if ( stringToDelete )
			free( stringToDelete );
	}
	else
	{
		long token_pos;
		theToken = find_token_by_pos( token_code_pos, &token_pos );
		if ( !theToken || token_pos )
		{
			/* this should never happen */
			theToken = NULL;
			fprintf(stderr, "Line %d # Could not find token for offset 0x%06lX\n", linenum, token_code_pos );
			printf("0x%03x \\ token not found\n", definedFCodeNumber);
		}
		else if ( good_token_name )
			printf( "0x%03x %s\n", definedFCodeNumber, theToken->name );
		else if (len && len <= 32) /* had length but was bad */
			printf( "0x%03x %s \\ «%s» bad name, unnamed\n", definedFCodeNumber, stringFromFile, theToken->name );
		else
			printf( "0x%03x \\ %s\n", definedFCodeNumber, theToken->name );

	
/* dump special token info */
		if ( theToken )
		{
			if ( mac_rom_fcode_flags & ( tokenFlagNoHeader | tokenFlagVisible | tokenFlagInvisible ) )
				fprintf(stderr, "Line %d # Defined %s%s%stoken 0x%03x \"%s\"\n", linenum, 
					(mac_rom_fcode_flags & tokenFlagNoHeader) ? "headerless " : "named ",
					(mac_rom_fcode_flags & tokenFlagInvisible) ? "invisible " : "",
					(mac_rom_fcode_flags & tokenFlagVisible) ? "visible " : "",
					definedFCodeNumber, theToken->name);

			if ( mac_rom_fcode_flags & tokenFlagImmediate )
				fprintf(stderr, "Line %d # Defined immediate token 0x%03x \"%s\"\n", linenum, definedFCodeNumber, theToken->name );

			if ( mac_rom_fcode_flags & tokenFlagInstance )
				fprintf(stderr, "Line %d # Defined instance token 0x%03x \"%s\"\n", linenum, definedFCodeNumber, theToken->name);

			if ( mac_rom_fcode_flags & tokenFlagVectored )
				fprintf(stderr, "Line %d # Defined vectored token 0x%03x \"%s\"\n", linenum, definedFCodeNumber, theToken->name);

			if ( mac_rom_fcode_flags & tokenFlagAlias )
				fprintf(stderr, "Line %d # Defined alias token 0x%03x \"%s\" to 0x%03x \"%s\"\n", linenum, definedFCodeNumber, theToken->name, aliasToken->fcode, aliasToken->name );
		}
		else
			printf( "0x%03x\n", definedFCodeNumber );
	}

	set_streampos( token_code_pos );

	mark_last_streampos();
}

static void bquote(void)
{
	s8 *string;
	u8 len;
	s8 strbuff[2000];

	string=get_string(&len);
	if ( gPass )
	{
		output_token();
		pretty_string(string, len, strbuff);
		printf("%s\n", strbuff);
	}
	free(string);
}

static void blit(void)
{
	u32 lit;

	lit=get_num32();
	if ( gPass )
	{
		output_token();
		printf("0x%lx\n",lit);
	}
}

static void offset16(void)
{
	decode_default();
	offs16=TRUE;
}

static void decode_branch(void)
{
	s16 offs;
	offs=decode_offset();
	if (offs>=0)
		indent++;
	else
		indent--;
}

static void decode_two(void)
{
	u16 token;

	output_token();
	token=get_token();
	decode_default();
}

static void decode_start(void)
{
	u8  fcformat;
	u16 fcchecksum, checksum=0;
	long newLen;
	long newEndPos;
	long numCheckSumBytes=0;
	long i;
	bool problem=TRUE;
	bool had_start=got_start;

	/* current_token_pos */			/* 0 */
	fcformat=get_num8();			/* 1 */
	fcchecksum=get_num16();			/* 2 */
	newLen=(long)get_num32();		/* 4 */
									/* 8 */

	if ( newLen > 8 )
	{
		newEndPos = newLen + current_token_pos;
		if (newEndPos >= newLen) /* no overflow */
		if (newEndPos <= filelen)
		{
			numCheckSumBytes = newLen - 8;
			
			for (i=0; i < numCheckSumBytes; i++)
				checksum+=get_num8();

			if (current_token_pos==0 || fcchecksum==checksum)
			{
				fclen=newEndPos;
				indent=0;
				problem=FALSE;

				offs16 = current_token_fcode != 0x0fd; /* version1 */
				gStartPos=current_token_pos;
				got_start = TRUE;
			}
		}
	}

	if (got_start)
	{
		isSpecialStart = FALSE;
		
		if (!had_start)
		{
			set_streampos(current_token_pos);
			decode_lines();
			if ( gPass )
				printf ("\\ detokenizing start at offset %06lX\n", current_token_pos + romstartoffset);
			dump_line( current_token_pos );
		}

		if ( gPass )
/*
		if (fcchecksum==checksum)
*/
			if (current_token_pos>0)
				fprintf(stderr, "Line %d # Start\n", linenum);

		decode_indent();
		decode_default();
		set_streampos(current_token_pos+1);

		dump_line(get_streampos());
		decode_indent();
		get_num8();
		if ( gPass )
			printf("  format:    0x%02x\n", fcformat);

		dump_line(get_streampos());
		decode_indent();
		get_num16();
		if ( gPass )
			printf("  checksum:  0x%04x (%s)\n", fcchecksum, numCheckSumBytes>0? (fcchecksum==checksum? "ok":"not ok"):"none");

		dump_line(get_streampos());
		decode_indent();
		get_num32();
		if ( gPass ) {
			if (newEndPos<=filelen)
				printf("  len:       0x%lx (%ld bytes)\n", newLen, newLen);
			else
				printf("  len:       0x%lx (%ld bytes) is greater than file length 0x%lx (%ld bytes)\n", newLen, newLen, filelen-current_token_pos, filelen-current_token_pos);
		}

		if (problem)
			set_streampos(current_token_pos+1);

		mark_last_streampos();
	}
	else 
		set_streampos(current_token_pos+1);
}



static void decode_rom_token2(void)
{
	if (current_token_pos > 3)
	if ( ( ( current_token_pos + 1 ) & 3 ) == 0 )
	if (current_token_pos - gLastPos < 32)
	{
		set_streampos( current_token_pos - 3 );
		if ( get_num32() == 0x10 )
		{
/*
			fclen=newEndPos;
*/
			indent=0;
			offs16 = FALSE;
			gStartPos=current_token_pos;
			got_start = TRUE;
			isSpecialStart = TRUE;
			decode_lines();
			if ( gPass )
				printf ("\\ detokenizing start at offset %06lX\n", current_token_pos+1 + romstartoffset);
			mark_last_streampos();
			if ( !gPass )
				fprintf(stderr, "Line %d # Unknown start\n", linenum);
		}
		else
			set_streampos(current_token_pos+1);
	}
}



static void local_variables_declaration(void)
{
	/* local variables { local_0 ... local_n-1 ; ... } */
	#warning "this needs to be tested with Open Firmware 1.0.5"

	int num=get_num8();
	if ( gPass )
	{
		int i;
		printf( "{ " );
		for ( i=0; i<num; i++ )
			printf( "local_%d ", i );
		printf( "; ... }" );
	}
	decode_default();
}



static void decode_token(u16 token_to_decode)
{
/*
	if ( get_streampos() >= 0x01F714 )
		DebugStr( "\phello" );
*/
	switch (token_to_decode) {
		case 0x0f0: /* start0 */
		case 0x0f1: /* start1 */
		case 0x0f2: /* start2 */
		case 0x0f3: /* start4 */
		case 0x0fd: /* version1 */
			decode_start();
			break;

		default:
		{
			if (got_start)
			{
				decode_indent();
				switch (token_to_decode)
				{
					case 0x000: /* end0 */
						decode_default();
						if ( isSpecialStart )
						{
							got_start = FALSE;
							isSpecialStart = FALSE;
							dump_line(get_streampos());
							if ( gPass )
								printf ("\\ detokenizing finished after %ld bytes.\n",
										get_streampos() - gStartPos );
						}
						break;
					case 0x0b5: /* new-token */
						new_token();
						break;
					case 0x0b6: /* named-token */
						named_token(FALSE);
						break;
					case 0x0ca: /* external-token */
						named_token(TRUE);
						break;
					case 0x012: /* b(") */
						bquote();
						break;
					case 0x010: /* b(lit) */
						blit();
						break;
					case 0x0cc: /* offset16 */
						offset16();
						break;
					case 0x013: /* bbranch */
					case 0x014: /* b?branch */
						decode_branch();
						break;
					case 0x0b7: /* b(:) */
					case 0x0b1: /* b(<mark) */
					case 0x0c4: /* b(case) */
						decode_default();
						indent++;
						break;
					case 0x0c2: /* b(;) */
					case 0x0b2: /* b(>resolve) */
					case 0x0c5: /* b(endcase) */
						decode_default();
						indent--;
						break;
					case 0x015: /* b(loop) */
					case 0x016: /* b(+loop) */
					case 0x0c6: /* b(endof) */
						decode_offset();
						indent--;
						break;
					case 0x017: /* b(do) */
					case 0x018: /* b/?do) */
					case 0x01c: /* b(of) */
						decode_offset();
						indent++;
						break;
					case 0x011: /* b(') */
					case 0x0c3: /* b(to) */
						decode_two();
						break;

/* PowerMac OF 1.0.5 ROM tokens */

	/*

	Forth:			{ local1, local2, ..., localn ; localn + 1, ... } \ ";" and local names are optional

	Tokenized:		0x401 n					\\ n is a byte (PMac OF 1.0.5 Mac rom max is 5)

	Compiled:		li r3,n
					bl mac_rom_code_45C

	Detokenized:	byte ^-7F61F8

	*/
					case 0x401:
						if (mac_rom & 1)
							local_variables_declaration();
						else
							goto thedefault;
						break;

					default:
thedefault:
						decode_default();
				} /* switch token_to_decode */
				mark_last_streampos();
			}
			else
			{
				switch (token_to_decode) {
					case 0x0bf: /* b(code) */
					case 0x0b7: /* b(:) */
					case 0x0b8: /* b(value) */
					case 0x0b9: /* b(variable) */
					case 0x0ba: /* b(constant) */
					case 0x0bb: /* b(create) */ /* •••••••••• */
					case 0x0bc: /* b(defer) */
					case 0x0bd: /* b(buffer:) */
					case 0x0be: /* b(field) */
					case 0x0db: /* set-token */
						decode_rom_token();
						break;
					case 0x010:
						if ( mac_rom & 2 )
							decode_rom_token2();
						break;
				}
			}
		} /* default */
	} /* switch */

	if ( got_start && current_token_record )
	if ( gPass )
	{
		if ( (current_token_record->flags & kFCodeTypeMask) == kFCodeHistorical )
			fprintf(stderr, "Line %d # Historical or non-implemented FCode: \"%s\" [0x%03x]\n", linenum, lookup_token(token_to_decode), token_to_decode);

		if ( (current_token_record->flags & kFCodeTypeMask) == kFCodeNew )
			fprintf(stderr, "Line %d # New FCode: \"%s\" [0x%03x]\n", linenum, lookup_token(token_to_decode), token_to_decode);
	}
}


int detokenize(void)
{
	for ( gPass = 0; gPass <= 1; gPass++ )
	{
		fprintf(stderr, "Doing Pass %d\n", gPass + 1);
		
		offs16=TRUE;
		isSpecialStart=FALSE;
		gStartPos=0;
		gLastPos=0;
		gLastMacRomTokenPos=-1;
		linenum=0;
		set_streampos(0);

		do {
			got_start = !(mac_rom != 0);

			do {
				if (got_start)
					decode_lines();

				current_token_pos = get_streampos();
				current_token_fcode = get_token();
				current_token_record = find_token( current_token_fcode );
				current_token_name = get_token_name( current_token_record );

				decode_token(current_token_fcode);

			} while ( (current_token_fcode||decode_all) && (get_streampos() < fclen) );

			decode_lines();
			

			if (gStartPos > 0)
			{
				if ( gPass )
				{
					printf ("\\ detokenizing finished after %ld of %ld bytes.\n",
							get_streampos() - gStartPos, fclen - gStartPos );
					fprintf(stderr, "Line %d # End\n", linenum);
				}
				gStartPos=0;
			}
			else
			{
				if ( gPass )
					printf ("\\ detokenizing finished after %ld of %ld bytes.\n",
							get_streampos(), fclen );
			}

			fclen = filelen;
			
		} while (ignore_len && (get_streampos() < filelen) );
	}

	if (mac_rom)
		dump_defined();

	return 0;
}
